package org.apache.bcel.util;

import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.IINC;
import org.apache.bcel.generic.LocalVariableInstruction;
import org.apache.bcel.generic.LDC2_W;
import org.apache.bcel.generic.LDC;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.InstructionConst;
import org.apache.bcel.generic.INSTANCEOF;
import org.apache.bcel.Const;
import org.apache.bcel.generic.FieldInstruction;
import org.apache.bcel.generic.ConstantPushInstruction;
import org.apache.bcel.generic.CHECKCAST;
import java.util.Locale;
import org.apache.bcel.generic.ArrayInstruction;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.MULTIANEWARRAY;
import org.apache.bcel.generic.NEWARRAY;
import org.apache.bcel.generic.CPInstruction;
import org.apache.bcel.generic.AllocationInstruction;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.Select;
import org.apache.bcel.generic.BranchHandle;
import org.apache.bcel.generic.Visitor;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.classfile.Utility;
import java.util.ArrayList;
import java.util.HashMap;
import org.apache.bcel.generic.BranchInstruction;
import java.util.List;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.Instruction;
import java.util.Map;
import org.apache.bcel.generic.ConstantPoolGen;
import java.io.PrintWriter;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.EmptyVisitor;

class BCELFactory extends EmptyVisitor
{
    private static final String CONSTANT_PREFIX;
    private final MethodGen methodGen;
    private final PrintWriter printWriter;
    private final ConstantPoolGen constantPoolGen;
    private final Map<Instruction, InstructionHandle> branchMap;
    private final List<BranchInstruction> branches;
    
    BCELFactory(final MethodGen mg, final PrintWriter out) {
        this.branchMap = new HashMap<Instruction, InstructionHandle>();
        this.branches = new ArrayList<BranchInstruction>();
        this.methodGen = mg;
        this.constantPoolGen = mg.getConstantPool();
        this.printWriter = out;
    }
    
    private void createConstant(final Object value) {
        String embed = value.toString();
        if (value instanceof String) {
            embed = '\"' + Utility.convertString(embed) + '\"';
        }
        else if (value instanceof Character) {
            embed = "(char)0x" + Integer.toHexString((char)value);
        }
        else if (value instanceof Float) {
            embed += "f";
        }
        else if (value instanceof Long) {
            embed += "L";
        }
        else if (value instanceof ObjectType) {
            final ObjectType ot = (ObjectType)value;
            embed = "new ObjectType(\"" + ot.getClassName() + "\")";
        }
        this.printWriter.println("il.append(new PUSH(_cp, " + embed + "));");
    }
    
    public void start() {
        if (!this.methodGen.isAbstract() && !this.methodGen.isNative()) {
            for (InstructionHandle ih = this.methodGen.getInstructionList().getStart(); ih != null; ih = ih.getNext()) {
                final Instruction i = ih.getInstruction();
                if (i instanceof BranchInstruction) {
                    this.branchMap.put(i, ih);
                }
                if (ih.hasTargeters()) {
                    if (i instanceof BranchInstruction) {
                        this.printWriter.println("    InstructionHandle ih_" + ih.getPosition() + ";");
                    }
                    else {
                        this.printWriter.print("    InstructionHandle ih_" + ih.getPosition() + " = ");
                    }
                }
                else {
                    this.printWriter.print("    ");
                }
                if (!this.visitInstruction(i)) {
                    i.accept(this);
                }
            }
            this.updateBranchTargets();
            this.updateExceptionHandlers();
        }
    }
    
    private void updateBranchTargets() {
        this.branches.forEach(bi -> {
            final BranchHandle bh = this.branchMap.get(bi);
            final int pos = bh.getPosition();
            final String name = bi.getName() + "_" + pos;
            final int targetPos = bh.getTarget().getPosition();
            this.printWriter.println("    " + name + ".setTarget(ih_" + targetPos + ");");
            if (bi instanceof Select) {
                final InstructionHandle[] ihs = ((Select)bi).getTargets();
                for (int j = 0; j < ihs.length; ++j) {
                    final int targetPos2 = ihs[j].getPosition();
                    this.printWriter.println("    " + name + ".setTarget(" + j + ", ih_" + targetPos2 + ");");
                }
            }
        });
    }
    
    private void updateExceptionHandlers() {
        final CodeExceptionGen[] exceptionHandlers;
        final CodeExceptionGen[] handlers = exceptionHandlers = this.methodGen.getExceptionHandlers();
        for (final CodeExceptionGen h : exceptionHandlers) {
            final String type = (h.getCatchType() == null) ? "null" : BCELifier.printType(h.getCatchType());
            this.printWriter.println("    method.addExceptionHandler(ih_" + h.getStartPC().getPosition() + ", ih_" + h.getEndPC().getPosition() + ", ih_" + h.getHandlerPC().getPosition() + ", " + type + ");");
        }
    }
    
    @Override
    public void visitAllocationInstruction(final AllocationInstruction i) {
        Type type;
        if (i instanceof CPInstruction) {
            type = ((CPInstruction)i).getType(this.constantPoolGen);
        }
        else {
            type = ((NEWARRAY)i).getType();
        }
        final short opcode = ((Instruction)i).getOpcode();
        int dim = 1;
        switch (opcode) {
            case 187: {
                this.printWriter.println("il.append(_factory.createNew(\"" + ((ObjectType)type).getClassName() + "\"));");
                break;
            }
            case 197: {
                dim = ((MULTIANEWARRAY)i).getDimensions();
            }
            case 188:
            case 189: {
                if (type instanceof ArrayType) {
                    type = ((ArrayType)type).getBasicType();
                }
                this.printWriter.println("il.append(_factory.createNewArray(" + BCELifier.printType(type) + ", (short) " + dim + "));");
                break;
            }
            default: {
                throw new IllegalArgumentException("Unhandled opcode: " + opcode);
            }
        }
    }
    
    @Override
    public void visitArrayInstruction(final ArrayInstruction i) {
        final short opcode = i.getOpcode();
        final Type type = i.getType(this.constantPoolGen);
        final String kind = (opcode < 79) ? "Load" : "Store";
        this.printWriter.println("il.append(_factory.createArray" + kind + "(" + BCELifier.printType(type) + "));");
    }
    
    @Override
    public void visitBranchInstruction(final BranchInstruction bi) {
        final BranchHandle bh = this.branchMap.get(bi);
        final int pos = bh.getPosition();
        final String name = bi.getName() + "_" + pos;
        if (bi instanceof Select) {
            final Select s = (Select)bi;
            this.branches.add(bi);
            final StringBuilder args = new StringBuilder("new int[] { ");
            final int[] matchs = s.getMatchs();
            for (int i = 0; i < matchs.length; ++i) {
                args.append(matchs[i]);
                if (i < matchs.length - 1) {
                    args.append(", ");
                }
            }
            args.append(" }");
            this.printWriter.print("Select " + name + " = new " + bi.getName().toUpperCase(Locale.ENGLISH) + "(" + (Object)args + ", new InstructionHandle[] { ");
            for (int i = 0; i < matchs.length; ++i) {
                this.printWriter.print("null");
                if (i < matchs.length - 1) {
                    this.printWriter.print(", ");
                }
            }
            this.printWriter.println(" }, null);");
        }
        else {
            final int tPos = bh.getTarget().getPosition();
            String target;
            if (pos > tPos) {
                target = "ih_" + tPos;
            }
            else {
                this.branches.add(bi);
                target = "null";
            }
            this.printWriter.println("    BranchInstruction " + name + " = _factory.createBranchInstruction(" + BCELFactory.CONSTANT_PREFIX + bi.getName().toUpperCase(Locale.ENGLISH) + ", " + target + ");");
        }
        if (bh.hasTargeters()) {
            this.printWriter.println("    ih_" + pos + " = il.append(" + name + ");");
        }
        else {
            this.printWriter.println("    il.append(" + name + ");");
        }
    }
    
    @Override
    public void visitCHECKCAST(final CHECKCAST i) {
        final Type type = i.getType(this.constantPoolGen);
        this.printWriter.println("il.append(_factory.createCheckCast(" + BCELifier.printType(type) + "));");
    }
    
    @Override
    public void visitConstantPushInstruction(final ConstantPushInstruction i) {
        this.createConstant(i.getValue());
    }
    
    @Override
    public void visitFieldInstruction(final FieldInstruction i) {
        final short opcode = i.getOpcode();
        final String className = i.getClassName(this.constantPoolGen);
        final String fieldName = i.getFieldName(this.constantPoolGen);
        final Type type = i.getFieldType(this.constantPoolGen);
        this.printWriter.println("il.append(_factory.createFieldAccess(\"" + className + "\", \"" + fieldName + "\", " + BCELifier.printType(type) + ", " + BCELFactory.CONSTANT_PREFIX + Const.getOpcodeName(opcode).toUpperCase(Locale.ENGLISH) + "));");
    }
    
    @Override
    public void visitINSTANCEOF(final INSTANCEOF i) {
        final Type type = i.getType(this.constantPoolGen);
        this.printWriter.println("il.append(new INSTANCEOF(_cp.addClass(" + BCELifier.printType(type) + ")));");
    }
    
    private boolean visitInstruction(final Instruction i) {
        final short opcode = i.getOpcode();
        if (InstructionConst.getInstruction(opcode) != null && !(i instanceof ConstantPushInstruction) && !(i instanceof ReturnInstruction)) {
            this.printWriter.println("il.append(InstructionConst." + i.getName().toUpperCase(Locale.ENGLISH) + ");");
            return true;
        }
        return false;
    }
    
    @Override
    public void visitInvokeInstruction(final InvokeInstruction i) {
        final short opcode = i.getOpcode();
        final String className = i.getClassName(this.constantPoolGen);
        final String methodName = i.getMethodName(this.constantPoolGen);
        final Type type = i.getReturnType(this.constantPoolGen);
        final Type[] argTypes = i.getArgumentTypes(this.constantPoolGen);
        this.printWriter.println("il.append(_factory.createInvoke(\"" + className + "\", \"" + methodName + "\", " + BCELifier.printType(type) + ", " + BCELifier.printArgumentTypes(argTypes) + ", " + BCELFactory.CONSTANT_PREFIX + Const.getOpcodeName(opcode).toUpperCase(Locale.ENGLISH) + "));");
    }
    
    @Override
    public void visitLDC(final LDC i) {
        this.createConstant(i.getValue(this.constantPoolGen));
    }
    
    @Override
    public void visitLDC2_W(final LDC2_W i) {
        this.createConstant(i.getValue(this.constantPoolGen));
    }
    
    @Override
    public void visitLocalVariableInstruction(final LocalVariableInstruction i) {
        final short opcode = i.getOpcode();
        final Type type = i.getType(this.constantPoolGen);
        if (opcode == 132) {
            this.printWriter.println("il.append(new IINC(" + i.getIndex() + ", " + ((IINC)i).getIncrement() + "));");
        }
        else {
            final String kind = (opcode < 54) ? "Load" : "Store";
            this.printWriter.println("il.append(_factory.create" + kind + "(" + BCELifier.printType(type) + ", " + i.getIndex() + "));");
        }
    }
    
    @Override
    public void visitRET(final RET i) {
        this.printWriter.println("il.append(new RET(" + i.getIndex() + ")));");
    }
    
    @Override
    public void visitReturnInstruction(final ReturnInstruction i) {
        final Type type = i.getType(this.constantPoolGen);
        this.printWriter.println("il.append(_factory.createReturn(" + BCELifier.printType(type) + "));");
    }
    
    static {
        CONSTANT_PREFIX = Const.class.getSimpleName() + ".";
    }
}
